// Copyright (c) 2007, 2008, 2009 libmv authors.
// Copyright (c) 2015 Julius Ziegler <ziegler@atlatec.de>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//
// A simple implementation of levenberg marquardt.
//
// [1] K. Madsen, H. Nielsen, O. Tingleoff. Methods for Non-linear Least
// Squares Problems.
// http://www2.imm.dtu.dk/pubdb/views/edoc_download.php/3215/pdf/imm3215.pdf
//
// TODO(keir): Cite the Lourakis' dogleg paper.

#ifndef LIBMV_NUMERIC_LEVENBERG_MARQUARDT_H
#define LIBMV_NUMERIC_LEVENBERG_MARQUARDT_H

#include <cmath>

#include <ceres/internal/autodiff.h>

#include "numeric.h"
#include "logging.h"

namespace libmv {
  template<typename AutoDiffableFunctor>
  class LevenbergMarquardt {
  public:
    typedef typename AutoDiffableFunctor::ResidualVector  ResidualVector;

    typedef typename AutoDiffableFunctor::ParameterVector ParameterVector;

    typedef typename ParameterVector::RealScalar          Scalar;

    typedef Eigen::PartialPivLU<
    Matrix<Scalar,
    ParameterVector::RowsAtCompileTime,
    ParameterVector::RowsAtCompileTime> > LinearSolver;

    typedef Matrix<Scalar,
    ResidualVector::RowsAtCompileTime,
    ParameterVector::RowsAtCompileTime>   JacobianMatrix;

    typedef Matrix<Scalar,
    JacobianMatrix::ColsAtCompileTime,
    JacobianMatrix::ColsAtCompileTime>    AMatrix;

    // TODO(keir): Some of these knobs can be derived from each other and
    // removed, instead of requiring the user to set them.
    enum Status {
      RUNNING,
      GRADIENT_TOO_SMALL,            // eps > max(J'*f(x))
      RELATIVE_STEP_SIZE_TOO_SMALL,  // eps > ||dx|| / ||x||
      ERROR_TOO_SMALL,               // eps > ||f(x)||
      HIT_MAX_ITERATIONS,
    };

    LevenbergMarquardt( const AutoDiffableFunctor &f )
      : f_(f) {}

    struct SolverParameters {
      SolverParameters()
        : gradient_threshold(1e-9),
          relative_step_threshold(1e-9),
          error_threshold(1e-9),
          initial_scale_factor(1e-3),
          max_iterations(100) {}
      Scalar gradient_threshold;       // eps > max(J'*f(x))
      Scalar relative_step_threshold;  // eps > ||dx|| / ||x||
      Scalar error_threshold;          // eps > ||f(x)||
      Scalar initial_scale_factor;     // Initial u for solving normal equations.
      int    max_iterations;           // Maximum number of solver iterations.
    };

    struct Results {
      Scalar error_magnitude;     // ||f(x)||
      Scalar gradient_magnitude;  // ||J'f(x)||
      int    iterations;
      Status status;
    };

    Status Update( const ParameterVector &x, const SolverParameters &params,
                   JacobianMatrix *J, AMatrix *A, ResidualVector *error, ParameterVector *g ) {

      // compute Jacobian with jets
      {
        static_assert( ParameterVector::RowsAtCompileTime != Eigen::Dynamic, "input (parameter vector-) size of problem must be constant" );

        typedef ceres::Jet<Scalar, ParameterVector::RowsAtCompileTime> Jet;

        std::vector<Jet> x_as_jets;
        x_as_jets.reserve( x.size() );

        // create "seeds" for single input variables
        for( int i=0; i<x.size(); i++ )
          {
            x_as_jets.push_back( Jet(x[i], i ) );
          }

        Jet f_as_jets[error->size()];

        // evaluate function
        f_( x_as_jets.data(), f_as_jets );

        // extract Jacobian and function value (error/residuals)
        for( int i=0; i<error->size(); i++ )
          {

            for( int j=0; j<x.size(); j++ )
              {
                (*J)(i,j) = f_as_jets[i].v[j];
              }
            (*error)[i] = -f_as_jets[i].a;
          }
      }

      // *J = df_(x);

      *A = (*J).transpose() * (*J);

      // *error = -f_(x);

      *g = (*J).transpose() * *error;
      if (g->array().abs().maxCoeff() < params.gradient_threshold) {
          return GRADIENT_TOO_SMALL;
        } else if (error->norm() < params.error_threshold) {
          return ERROR_TOO_SMALL;
        }
      return RUNNING;
    }

    Results minimize(ParameterVector *x_and_min) {
      SolverParameters params;
      return minimize(params, x_and_min);
    }

    Results minimize(const SolverParameters &params, ParameterVector *x_and_min) {
      ParameterVector &x = *x_and_min;
      JacobianMatrix J;
      AMatrix A;
      ResidualVector error;
      ParameterVector g;

      Results results;
      results.status = Update(x, params, &J, &A, &error, &g);

      Scalar u = Scalar(params.initial_scale_factor*A.diagonal().maxCoeff());
      Scalar v = 2;

      ParameterVector dx, x_new;
      int i;
      for (i = 0; results.status == RUNNING && i < params.max_iterations; ++i) {
          // TODO: this evaluation is not necessary
          ResidualVector f;
          f_( x.data(), f.data() );

          VLOG(3) << "iteration: " << i;
          VLOG(3) << "||f(x)||: " << f.norm();
          VLOG(3) << "max(g): " << g.array().abs().maxCoeff();
          VLOG(3) << "u: " << u;
          VLOG(3) << "v: " << v;

          AMatrix A_augmented = A + u*AMatrix::Identity(J.cols(), J.cols());
          LinearSolver solver(A_augmented);
          dx = solver.solve(g);
          bool solved = (A_augmented * dx).isApprox(g);
          if (!solved) {
              LOG(ERROR) << "Failed to solve";
            }
          if (solved && dx.norm() <= params.relative_step_threshold * x.norm()) {
              results.status = RELATIVE_STEP_SIZE_TOO_SMALL;
              break;
            }
          if (solved) {
              x_new = x + dx;
              // Rho is the ratio of the actual reduction in error to the reduction
              // in error that would be obtained if the problem was linear.
              // See [1] for details.

              ResidualVector f_new;
              f_( x_new.data(), f_new.data() );

              Scalar rho((error.squaredNorm() - f_new.squaredNorm())
                         / dx.dot(u*dx + g));
              if (rho > 0) {
                  // Accept the Gauss-Newton step because the linear model fits well.
                  x = x_new;
                  results.status = Update(x, params, &J, &A, &error, &g);
                  Scalar tmp = Scalar(2*rho-1);
                  u = u*std::max(1/3., 1 - (tmp*tmp*tmp));
                  v = 2;
                  continue;
                }
            }
          // Reject the update because either the normal equations failed to solve
          // or the local linear model was not good (rho < 0). Instead, increase u
          // to move closer to gradient descent.
          u *= v;
          v *= 2;
        }
      if (results.status == RUNNING) {
          results.status = HIT_MAX_ITERATIONS;
        }
      results.error_magnitude = error.norm();
      results.gradient_magnitude = g.norm();
      results.iterations = i;
      return results;
    }

  private:
    const AutoDiffableFunctor &f_;
  };

}  // namespace mv

#endif  // LIBMV_NUMERIC_LEVENBERG_MARQUARDT_H
